from vinca.lib import ansi
from vinca.lib.terminal import COLUMNS
from vinca.lib.fancy_input import fancy_input

def edit(card):
	front_path = (card.path / 'front')
	back_path  = (card.path / 'back')
	front = front_path.read_text()
	back  =  back_path.read_text()

	# TODO multiline fancy_input using Alt+Enter
	new_front = fancy_input(text = front, prompt = 'Q: ')
	front_path.write_text(new_front)
	
	new_back = fancy_input(text = back, prompt = 'A: ')
	back_path.write_text(new_back)




import datetime, time
import functools
TODAY = datetime.date.today()

# want to express this as a generator to be called in the individual generator files
def generator(g):
	# prserve wrapped functions name, docstring, and signature
	@functools.wraps(g)
	def wrapped_generator():
		start = time.time()
		new_cards = g()
		stop = time.time()
		elapsed = int(stop - start)
		for card in new_cards:
			card.make_string()
			card.add_history(TODAY, elapsed, 'create')
			card.load_metadata()  # TODO why do I need this?
		return new_cards[0]
	return wrapped_generator



'''Vinca 112
Simple Spaced Repetition'''

from functools import partial as _partial
from pathlib import Path as _Path
from vinca.cardlist import Cardlist as _Cardlist
from vinca.card import Card as _Card
import fire as _fire
# generators
from vinca.generators.two_lines import two_lines
q = two_lines
from vinca.generators.verses import verses
v = verses
from vinca.generators.media import media
m = media

_cards_path = _Path.home() / 'cards'
_ALL_CARDS = [_Card(int(id.name)) for id in _cards_path.iterdir()] 

col = _Cardlist(_ALL_CARDS)

# methods of the collection
browse = b = col.browse
filter = f =  col.filter
sort = col.sort
review = r =  col.review
save = col.save
load = col.load
purge = col.purge

help = lambda: 'Use --help instead'
version = lambda: 112

_fire.Fire()
'''
Add the following code to the ActionGroup object in helptext.py of fire to get proper aliasing
A better way would be to go back further into the code and check if two functions share the same id

  def Add(self, name, member=None):
    if member and member in self.members:
      dupe = self.members.index(member)
      self.names[dupe] += ', ' + name
      return
    self.names.append(name)
    self.members.append(member)
'''
'''
Make this substitution on line 458 of core.py to allow other iterables to be accessed by index

    # is_sequence = isinstance(component, (list, tuple))
    is_sequence = hasattr(component, '__getitem__') and not hasattr(component, 'values')
'''
'''
And make a corresponding change in generating the help message

  is_sequence = hasattr(component, '__getitem__') and not hasattr(component, values)
  # if isinstance(component, (list, tuple)) and component:
  if is_sequence and component:
'''



from vinca.generators.media import media
from vinca.generators.two_lines import two_lines
from vinca.generators.verses import verses

generators_dict = {'m': media,
		   'v': verses,
		   '2': two_lines}



from vinca.card import Card
from vinca.generators.decorator import generator

@generator
def verses():
	''' make a card which quizzes one line at a time\nfor poetry, oratory, recipes '''
	new_card = Card(create=True)
	new_card.editor, new_card.reviewer, new_card.scheduler = 'verses', 'verses', 'base'
	new_card.edit()
	return [new_card]



from vinca.lib import ansi
import shutil

COLUMNS, LINES = shutil.get_terminal_size()

# A Context Manager for the terminal's alternate screen
class AlternateScreen:
	def __enter__(self):
		ansi.save_cursor()
		ansi.hide_cursor()
		ansi.save_screen()
		ansi.clear()
		ansi.move_to_top()

	def __exit__(self, *exception_args):
		ansi.restore_screen()
		ansi.restore_cursor()


class NoCursor:
	def __enter__(self):
		ansi.hide_cursor()

	def __exit__(self, *exception_args):
		ansi.show_cursor()
		

class LineWrapOff:
	def __enter__(self):
		ansi.line_wrap_off()

	def __exit__(self, *exception_args):
		ansi.line_wrap_on()

def count_screen_lines(text):
	text = ansi.strip_ansi(text)
	lines = text.splitlines()
	lines = [l.split('\r')[-1] for l in lines] # carriage return overwrites the line so we only want the text after the last carriage return on the line
	lines = [len(l) // COLUMNS for l in lines]
	return sum(lines)



# linear editor for lyrics, poetry, oratory, etc.
import subprocess
from pathlib import Path
import shutil

vinca_path = Path(__file__).parent.parent.parent # I counted right
tags_path = vinca_path / 'data' / 'tags.txt'

path = Path(__file__).parent
vimrc_path = path / 'vimrc'

def edit(card):
	# we are going to run vim...
	vim_cmd = ['vim']
	vim_cmd += [card.path/'lines']
	# specify custom vimrc
	vim_cmd += ['-Nu', vimrc_path]
	# launch
	subprocess.run(vim_cmd)



''' This scrollback module is deprecated
The idea was going to be to create a wrapper for a function which would intercept its print statements compute the number of terminal lines printed and scrollback that many lines and clear to bottom. This could be used, for example to show card editing or statistics alongside the browser. However we do have to use a time proxy for interactivity to know whether to continue
'''

import sys
import functools
from vinca.lib import ansi
from vinca.lib.terminal import count_screen_lines
from vinca.lib.readkey import readkey

class Intercepter:
	def __init__(self):
		self.text = ''
	def write(self, text):
		sys.__stdout__.write(text) # echo to terminal
		self.text += text # log the printed text
	def __enter__(self):
		sys.stdout = self # begin intercepting stdout
	def __exit__(self):
		sys.stdout = sys.__stdout__ # reconnect normal stdout

def scrollback(func):
	@functools.wraps(func)
	def modified_func(*args, scrollback = False, **kwargs):
		if not scrollback:
			return func(*args, **kwargs)
		with Intercepter() as i:
			ret = func(*args, **kwargs)
			n = count_screen_lines(i.text)
		ansi.up_line(n)
		ansi.clear_to_end()
		return ret
	return modified_func



import re
import datetime
from functools import partial
from shutil import copytree, rmtree
from pathlib import Path
TODAY = datetime.date.today()
DAY = datetime.timedelta(days = 1)
from vinca.browser import Browser
from vinca.generators import generators_dict
from vinca.lib.terminal import LineWrapOff
from vinca.lib import ansi
from vinca.lib.fancy_input import fancy_input
from vinca.lib.readkey import readkey

cards_path = Path('~/cards')

class Cardlist:
	''' this is a collection of cards. most of the user interface takes place through the browser '''

	def __init__(self, _cards):
		# we do not subclass list so that
		# the fire help is not littered
		# with inherited methods
		self._cards = _cards
		self._hotkeys = {'D': self.mass_delete,
					      'T': self.edit_tags}
		self._confirm_exit_commands = []

	def __iter__(self):
		return iter(self._cards)

	def __len__(self):
		return len(self._cards)

	def __getitem__(self, slice):
		return self._cards[slice]


	def __str__(self):
		s = ''
		l = len(self)
		if l == 0:
			return 'No cards.'
		if l > 10:
			s += f'10 of {l}\n'
		s += ansi.codes['line_wrap_off']
		for card in self[:10]:
			if card.due_as_of(TODAY):
				s += ansi.codes['bold']
				s += ansi.codes['blue']
			if card.deleted:
				s += ansi.codes['crossout']
				s += ansi.codes['red']
			s += f'{card.id}\t{card}\n'
			s += ansi.codes['reset']
		s += ansi.codes['line_wrap_on']
		return s

	def browse(self):
		''' scroll through your collection with j and k '''
		Browser(self).browse()

	def review(self):
		''' review all cards '''
		Browser(self).review()
				
	def mass_delete(self):
		for card in self:
			card.deleted = True

	def add_tag(self, tag):
		for card in self:
			card.tags += [tag]

	def remove_tag(self, tag):
		for card in self:
			if tag in card.tags:
				card.tags.remove(tag)
			# TODO do this with set removal
			card.save_metadata()

	def count(self):
		''' simple summary statistics '''
		total_count = len(self)
		new_count = len(self.filter(new_only=True))
		due_count = len(self.filter(due_only=True))
		print('total',total_count,sep='\t')
		print('new',new_count,sep='\t')
		print('due',due_count,sep='\t')

	def edit_tags(self):
		tags_add = fancy_input(prompt = 'tags to add: ', completions = ALL_TAGS).split()
		tags_remove = fancy_input(prompt = 'tags to remove: ', completions = ALL_TAGS).split()
		for tag in tags_add:
			self.add_tag(tag)
		for tag in tags_remove:
			self.remove_tag(tag)

	def save(self, save_path):
		''' backup your cards '''
		save_path = Path(save_path)
		for card in self:
			copytree(card.path, save_path / str(card.id))

	@staticmethod
	def load(load_path, overwrite = False):
		load_path = Path(load_path)
		if overwrite:
			rmtree(cards_path)
			copytree(load_path, cards_path)
			return
		old_ids = [card.id for card in ALL_CARDS]
		max_old_id = max(old_ids, default = 1)
		for new_id,card_path in enumerate(load_path.iterdir(), max_old_id + 1):
			copytree(card_path, cards_path / str(new_id))


	def purge(self):
		''' Permanently delete all cards marked for deletion. '''
		deleted_cards = self.filter(deleted_only = True)
		if not deleted_cards:
			print('no cards are marked for deletion.')
			return
		print(f'delete {len(deleted_cards)} cards? (y/n)')
		if (confirmation := readkey()) == 'y':
			for card in deleted_cards:
				rmtree(card.path)

	def delete(self):
		for card in self:
			card.delete(toggle = False)



	def filter(self, pattern='', 
		   tags_include={}, tags_exclude={}, # specify a SET of tags
		   create_date_min=None, create_date_max=None,
		   seen_date_min=None, seen_date_max=None,
		   due_date_min=None, due_date_max=None,
		   editor=None, reviewer=None, scheduler=None,
		   deleted_only=False, 
		   due_only=False,
		   new_only=False,
		   invert=False):
		''' try --due_only or --pattern='Gettysburg Address' '''
		
		# cast dates to datetime format
		for date in ('create_date_min','create_date_max','seen_date_min','seen_date_max','due_date_min','due_date_max'):
			arg = locals()[date]
			if arg is int:
				locals()[date] = TODAY + arg*DAY
			elif isinstance(arg, str):
				try:
					locals()[date] = datetime.date.fromisoformat(arg)
				except ValueError as e:
					print(f'Dates must be integer or iso format')
					raise e

		if due_only: due_date_max = TODAY
		# compile the regex pattern for faster searching
		p = re.compile(f'({pattern})')  # wrap in parens to create regex group \1

		tags_include, tags_exclude = set(tags_include), set(tags_exclude)

		f = lambda card: (((not tags_include or bool(tags_include & set(card.tags))) and
				(not tags_exclude or not bool(tags_exclude & set(card.tags))) and
				(not create_date_min or create_date_min <= card.create_date) and
				(not create_date_max or create_date_max >= card.create_date) and 
				(not seen_date_min or seen_date_min <= card.seen_date) and
				(not seen_date_max or seen_date_max >= card.seen_date) and 
				(not due_date_min or due_date_min <= card.due_date) and
				(not due_date_max or due_date_max >= card.due_date) and 
				(not editor or editor == card.editor) and
				(not reviewer or reviewer == card.reviewer) and
				(not scheduler or scheduler == card.scheduler) and
				(card.deleted or not deleted_only) and
				(not new_only or card.new) and
				(not pattern or bool(p.search(card.string)))) != 
				invert)
		
		# matches.sort(key=lambda card: card.seen_date, reverse=True)
		return self.__class__([c for c in self if f(c)])

	def sort(self, create_date=False, seen_date=False, due_date=False, reverse=False):
		''' sort the collection '''
		if create_date:
			return self.__class__(sorted(self,
				key=lambda card: card.create_date, reverse=not reverse))
		if seen_date:
			return self.__class__(sorted(self,
				key=lambda card: card.seen_date, reverse=not reverse))
		if due_date:
			return self.__class__(sorted(self,
				key=lambda card: card.due_date, reverse=reverse))
		print('supply a criterion: --create_date | --seen_date | --due_date')



from vinca.card import Card
from vinca.generators.decorator import generator

@generator
def media():
	''' make a card with audio and pictures '''
	# initialize card
	new_card = Card(create=True)
	# initialize card metadata
	new_card.editor, new_card.reviewer, new_card.scheduler = 'media', 'media', 'base'
	# initialize card data files
	for side in ('front', 'back'):
		(new_card.path / side).touch()
	new_card.edit()  
	return [new_card]
